#include "TemplateNodes.h"

#include "StringUtils.h"

TemplateNode::TemplateNode(TemplateNode* parent)
        : mParent(parent)
{

}

TemplateNode* TemplateNode::getParent() const
{
    return mParent;
}

void TemplateNode::addChild(std::unique_ptr<TemplateNode> child)
{
    mChildren.push_back(std::move(child));
}

std::size_t TemplateNode::getNumChildren() const
{
    return mChildren.size();
}

std::string TemplateNode::getIdentifier() const
{
    return {};
}

TemplateNode* TemplateNode::getChild(std::size_t index) const
{
    return mChildren[index].get();
}

std::string TemplateNode::getRawContent() const
{
    std::string content;
    for (const auto& child : mChildren)
    {
        content += child->getRawContent() + "\n";
    }
    return content;
}

void TemplateNode::setExtensionParent(TemplateNode* parent)
{
    mExtensionParent = parent;
}

std::string TemplateNode::render(TemplateNode* parentContext)
{
    if (!parentContext && mExtensionParent)
    {
        parentContext = mExtensionParent;
    }

    std::string content;
    for (size_t idx = 0; idx < mChildren.size(); idx++)
    {
        content += mChildren[idx]->render(parentContext);
    }
    return content;
}

TemplateExtends::TemplateExtends(TemplateNode* parent, const std::string& path)
    : TemplateNode(parent)
{
    mPath = StringUtils::stripQuotes(path);
};

std::string TemplateExtends::getRawContent() const
{
    return "TemplateExtends: " + mPath;
}

std::string TemplateExtends::getPath() const
{
    return mPath;
}

TemplateBlock::TemplateBlock(TemplateNode* parent, const std::string& name)
    : TemplateNode(parent),
    mName(name)
{

}

void TemplateBlock::addLine(const std::string& line)
{
    mBody.push_back(line);
}

std::string TemplateBlock::getRawContent() const
{
    std::string content = "TemplateBlock: " + mName + "\n";
    content += TemplateNode::getRawContent();
    return content;
}

std::string TemplateBlock::renderAsParent(TemplateNode* base)
{
    return {};
}

std::string TemplateBlock::render(TemplateNode* parentContext)
{
    std::string content;
    if (parentContext)
    {
        if (auto parent_node = parentContext->getFirstChildShallow<TemplateBlock>(getIdentifier()))
        {
            content = dynamic_cast<TemplateBlock*>(parent_node)->renderAsParent(this);
        }
    }
    else
    {

    }
    return content;
}

TemplateExpression::TemplateExpression(TemplateNode* parent, const std::string& content)
    : TemplateNode(parent),
    mContent(content)
{

}

std::string TemplateExpression::getRawContent() const
{
    return "TemplateExpression: " + mContent;
}

TemplateTextBody::TemplateTextBody(TemplateNode* parent)
        : TemplateNode(parent)
{

}

void TemplateTextBody::addLine(const std::string& content)
{
    mContent.push_back(content);
}

bool TemplateTextBody::hasContent() const
{
    return !mContent.empty();
}

std::string TemplateTextBody::render(TemplateNode* parentContext)
{
    std::string content;
    for (const auto& line : mContent)
    {
        content += line + '\n';
    }
    return content;
}

std::string TemplateTextBody::getRawContent() const
{
    std::string content;
    for (const auto& line : mContent)
    {
        content += "TemplateBody: " + line + "\n";
    }
    return content;
}
