#include "MarkdownParser.h"

#include "MarkdownDocument.h"
#include "StringUtils.h"

#include <sstream>
#include <iostream>

MarkdownParser::MarkdownParser()
{

}

MarkdownParser::~MarkdownParser()
{

}

void MarkdownParser::onMultilineQuote()
{
    std::cout << "Adding multiline quote " << mDocumentContent << std::endl;
    auto quote = std::make_unique<MarkdownMultilineQuote>(mWorkingTag);
    quote->appendTextContent(mDocumentContent);

    mDocumentContent.clear();
    mWorkingTag.clear();

    mDocumentState = DocumentState::NONE;
    mMarkdownDocument->addElement(std::move(quote));

    onNewParagraph();
}

void MarkdownParser::onInlineQuote()
{
    std::cout << "Adding inline quote " << mLineContent << std::endl;

    auto quote = std::make_unique<MarkdownInlineQuote>();
    quote->appendTextContent(mLineContent);
    mLineContent.clear();

    mLineState = LineState::NONE;
    if(mWorkingParagraph)
    {
        mWorkingParagraph->addChild(std::move(quote));
    }
}

void MarkdownParser::onHeading(unsigned level)
{
    std::cout << "Adding heading: " << mLineContent << std::endl;

    auto heading = std::make_unique<MarkdownHeading>(level);
    heading->appendTextContent(mLineContent);
    mMarkdownDocument->addElement(std::move(heading));
}

void MarkdownParser::onNewParagraph()
{
    if (mWorkingBulletList)
    {
        std::cout << "Adding bullets to document" << std::endl;
        mMarkdownDocument->addElement(std::move(mWorkingBulletList));
        mWorkingBulletList.reset();
        mDocumentState == DocumentState::NONE;
    }
    else if (mWorkingParagraph)
    {
        onTextSpan();

        if (!mWorkingParagraph->getNumChildren() == 0)
        {
            std::cout << "Adding para to document" << std::endl;
            mMarkdownDocument->addElement(std::move(mWorkingParagraph));
        }
    }
    mWorkingParagraph = std::make_unique<MarkdownParagraph>();
}

void MarkdownParser::onTextSpan()
{
    mLineContent.clear();

    if(mWorkingParagraph && !mDocumentContent.empty())
    {
        std::cout << "Adding text " << mDocumentContent << std::endl;

        auto text_span = std::make_unique<MarkdownTextSpan>();
        text_span->appendTextContent(mDocumentContent);
        mWorkingParagraph->addChild(std::move(text_span));
        mDocumentContent.clear();
    }
}

std::pair<unsigned, bool> MarkdownParser::onTick(unsigned tickCount)
{
    unsigned new_tick_count = tickCount;
    bool stop_line_processing = false;

    if (tickCount == 2)
    {
        if (mDocumentState == DocumentState::IN_MULTILINEQUOTE)
        {
            onMultilineQuote();
            stop_line_processing = true;
        }
        else
        {
            onNewParagraph();
            mLineState = LineState::IN_MULTILINE_TAG;
            new_tick_count = 0;
            mDocumentState = DocumentState::IN_MULTILINEQUOTE;
        }
    }
    else if(mLineState == LineState::IN_INLINEQUOTE)
    {
        if (mLineContent.empty())
        {
            mLineState = LineState::NONE;
            new_tick_count++;
        }
        else
        {
            new_tick_count = 0;
            onInlineQuote();
        }
    }
    else if(mDocumentState == DocumentState::IN_MULTILINEQUOTE)
    {
        new_tick_count++;
        mLineContent += '`';
    }
    else
    {
        new_tick_count++;
        mLineState = LineState::IN_INLINEQUOTE;
    }
    return {new_tick_count, stop_line_processing};
}

void MarkdownParser::onLink()
{
    std::cout << "Adding hyperlink to " << mLineContent << " with tag " << mWorkingTag << std::endl;
    auto element = std::make_unique<MarkdownLink>(mLineContent);
    mLineContent.clear();

    element->appendTextContent(mWorkingTag);
    mWorkingTag.clear();

    if (mWorkingParagraph)
    {
        mWorkingParagraph->addChild(std::move(element));
    }
    mLineState = LineState::NONE;
}

void MarkdownParser::onImage()
{
    std::cout << "Adding image with path " << mLineContent << " and alt" << mWorkingTag << std::endl;
    auto element = std::make_unique<MarkdownImage>(mLineContent, mWorkingTag);
    mLineContent.clear();

    element->appendTextContent(mWorkingTag);
    mWorkingTag.clear();

    if (mWorkingParagraph)
    {
        mWorkingParagraph->addChild(std::move(element));
    }
    mLineState = LineState::NONE;
}

void MarkdownParser::onBulletItem()
{
    std::cout << "Adding bullet item " << mLineContent << std::endl;

    if (!mWorkingBulletList)
    {
        mWorkingBulletList = std::make_unique<MarkdownBulletList>();
        mDocumentState == DocumentState::IN_BULLETS;
    }

    auto item = std::make_unique<MarkdownBulletItem>();
    item->appendTextContent(mLineContent);
    mLineContent.clear();

    mWorkingBulletList->addChild(std::move(item));
}

void MarkdownParser::processLine()
{
    mLineContent.clear();
    mLineState = LineState::NONE;

    unsigned heading_level{0};
    unsigned tick_count{0};
    bool flushed_pre_inline = false;

    bool first_nonspace = false;
    for(auto c : mWorkingLine)
    {
        if (!StringUtils::IsSpace(c))
        {
            if (first_nonspace)
            {
                first_nonspace = false;
            }
            else
            {
                first_nonspace = true;
            }
        }
        else
        {
            first_nonspace = false;
        }

        if (c == '`')
        {
            auto [ret_tick_count, stop_line_processing] = onTick(tick_count);
            tick_count = ret_tick_count;
            if(stop_line_processing)
            {
                return;
            }
        }
        else
        {
            if (mLineState == LineState::IN_INLINEQUOTE)
            {
                if (!flushed_pre_inline)
                {
                    std::cout << "Flushing pre-line " << std::endl;
                    mDocumentContent += mLineContent;
                    onTextSpan();
                    flushed_pre_inline = true;
                }
                mLineContent += c;
            }
            else if (mDocumentState == DocumentState::IN_MULTILINEQUOTE)
            {
                mLineContent += c;
            }
            else if(mLineState == LineState::IN_LINK_TAG)
            {
                if (c == ']')
                {
                    mLineState = LineState::AWAITING_LINK_BODY;
                }
                else
                {
                    mWorkingTag += c;
                }
            }
            else if(mLineState == LineState::AWAITING_LINK_BODY)
            {
                if (c == '(')
                {
                    mLineState = LineState::IN_LINK_BODY;
                }
                else
                {
                    mLineContent = '[' + mWorkingTag + ']';
                    mLineState = LineState::NONE;
                }
            }
            else if(mLineState == LineState::IN_LINK_BODY)
            {
                if(c==')')
                {
                    onLink();
                }
                else
                {
                    mLineContent += c;
                }
            }
            else if(mLineState == LineState::AWAITING_IMG_TAG)
            {
                if (c == '[')
                {
                    mLineState = LineState::IN_IMG_TAG;
                }
                else
                {
                    mLineContent = "![";
                    mLineState = LineState::NONE;
                }
            }
            else if(mLineState == LineState::IN_IMG_TAG)
            {
                if (c == ']')
                {
                    mLineState = LineState::AWAITING_IMG_BODY;
                }
                else
                {
                    mWorkingTag += c;
                }
            }
            else if(mLineState == LineState::AWAITING_IMG_BODY)
            {
                if (c == '(')
                {
                    mLineState = LineState::IN_IMG_BODY;
                }
                else
                {
                    mLineContent = "![" + mWorkingTag + "]";
                    mWorkingTag.clear();
                    mLineState = LineState::NONE;
                }
            }
            else if(mLineState == LineState::IN_IMG_BODY)
            {
                if (c == ')')
                {
                    onImage();
                }
                else
                {
                    mLineContent += c;
                }
            }
            else
            {
                if (c == '#')
                {
                    onNewParagraph();
                    mLineState = LineState::IN_HEADING;
                    heading_level++;
                }
                else if(c == '[')
                {
                    mDocumentContent += mLineContent;
                    onTextSpan();
                    mLineState = LineState::IN_LINK_TAG;
                }
                else if(c == '!')
                {
                    mDocumentContent += mLineContent;
                    onTextSpan();
                    mLineState = LineState::AWAITING_IMG_TAG;
                }
                else if(first_nonspace && c == '*')
                {
                    if (!mWorkingBulletList)
                    {
                        onNewParagraph();
                    }
                    mLineState = LineState::IN_BULLETS;
                }
                else
                {
                    mLineContent += c;
                }
            }
        }
    }

    if (mLineState == LineState::IN_HEADING)
    {
        onHeading(heading_level);
    }
    else if(mLineState == LineState::IN_MULTILINE_TAG)
    {
        mWorkingTag = mLineContent;
    }
    else if (mLineState == LineState::IN_INLINEQUOTE)
    {
        onTextSpan();
    }
    else if (mLineState == LineState::IN_BULLETS)
    {
        onBulletItem();
    }
    else
    {
        if (mLineContent.size() > 0)
        {
            mDocumentContent.append(mLineContent);
        }
    }
}

void MarkdownParser::onEmptyLine()
{
    onNewParagraph();
}

std::unique_ptr<MarkdownDocument> MarkdownParser::run(const std::string& content)
{
    mMarkdownDocument = std::make_unique<MarkdownDocument>();

    std::stringstream ss(content);
    std::string line;

    while (std::getline(ss, line, '\n'))
    {
        if (line.empty())
        {
            onEmptyLine();
            continue;
        }
        mWorkingLine = line;
        processLine();
    }

    onTextSpan();
    onNewParagraph();

    return std::move(mMarkdownDocument);
}
